import os
import re
import requests
from getpass import getpass
from pycodya import config
from pycodya.helpers import tools

class CodyaApi(object):
    def __init__(self, token_branch=""):
        self.token = self._extract_user_token()
        self.token_branch = token_branch if token_branch else self._extract_branch_token()
        self.base_url = "http://localhost:8080/api/v0"


    def _extract_user_token(self):
        if os.path.isfile(config.CREDS_GENERATED_FILE):
            with open(config.CREDS_GENERATED_FILE, 'r') as f:
                token = f.read()
            return token    
        else:
            return ""
    
    def _extract_branch_token(self):
        if os.path.isfile(config.BRANCH_TOKEN_FILE):
            with open(config.BRANCH_TOKEN_FILE, 'r') as f:
                branch_token = f.read()
            return branch_token    
        else:
            return ""


    def login(self):
        username = input("Username: ")
        password = getpass("Password: ")
        url = self.base_url + "/login"
        resp = requests.post(url, json={"username": username, "password": password})
        
        if resp.status_code == 200:    
            self.token = resp.json()["token"]
            
            if not os.path.isdir(config.DIR_GENERATED_ROOT):
                os.makedirs(config.DIR_GENERATED_ROOT)
            with open(config.CREDS_GENERATED_FILE, 'w') as f:
                f.write(self.token)
            print("login succesfull", username, password, self.token)
        else:
            print("Login unsuccessful", resp.text)

    def projects(self):
        token_branch = self._extract_branch_token()

        if not token_branch:
            print("No current project selected.. \n\n")
        else:
            url = self.base_url + "/branch-by-token"
            resp = requests.get(url, headers={'Authorization': 'bearer ' + token_branch})
            branch = resp.json()["branch"]
            project_id = branch["projectId"]
            url = self.base_url + "/project/" + project_id
            resp = requests.get(url, headers={'Authorization': 'bearer ' + self.token})
            if resp.status_code != 200:
                print("Please login first: pycodya login")
                return
            project = resp.json()["project"]
            print("You are currently on the project \033[1;3m{}\033[0m on branch \033[1;3m{}\033[0m\n\n".format(project["name"], branch["name"]))


        is_project_creation = input("What would you like to do:\n0 - Create a new project\n1 - Use an existing project\nPlease type 0 or 1: ")
        if is_project_creation != "0" and is_project_creation != "1":
            print("Wrong input: only 0 and 1 allowed") 
            return

        elif is_project_creation == "0":
            project_name = input("Please type the name of the project: ")
            if not bool(re.match("^[a-zA-Z0-9_]*$", project_name)): 
                print("Wrong input: no special characters please, only numbers letters and _ allowed" ) 
                return
            url = self.base_url + "/project"
            resp = requests.post(url, headers={'Authorization': 'bearer ' + self.token}, json={"name": project_name, "language": "pycodya"})
            if resp.status_code != 200:
                print("Please login first: pycodya login")
                return
            project_id = resp.json()["projectId"]
            url = self.base_url + "/project/" + project_id + "/branch"
            resp = requests.post(url, headers={'Authorization': 'bearer ' + self.token}, json={"name": "master"})
            branch_id = resp.json()["branchId"]
            url = self.base_url + "/branch/" + branch_id + "/token"
            resp = requests.get(url, headers={'Authorization': 'bearer ' + self.token})
            branch_token = resp.json()['token']
            with open(config.BRANCH_TOKEN_FILE, 'w') as f:
                f.write(branch_token)
            print("Project {} created with branch master. Please copy the token:\n{}".format(project_name, branch_token))
            return
        
        elif is_project_creation == "1":
            url = self.base_url + "/projects"
            resp = requests.get(url, headers={'Authorization': 'bearer ' + self.token})
            if resp.status_code != 200:
                print("Please login first: pycodya login")
                return
            all_projects = resp.json()["projects"]
            if not all_projects:
                print("No projects yet - please create one")
                return
            print("\nPlease see the list of all the projects:")
            for key, project in enumerate(all_projects):
                print("{} - {}".format(key, project["name"]))
            chosen_project = input("Please type a number to choose between the different projects: ")    
            if chosen_project.isnumeric() and int(chosen_project) >= 0 and int(chosen_project) < len(all_projects):
                project_id = all_projects[int(chosen_project)]["_id"]
                url = self.base_url + "/project/" + project_id + "/branches"
                resp = requests.get(url, headers={'Authorization': 'bearer ' + self.token})
                branches = resp.json()["branches"]
                chosen_branch = [b for b in branches if b["name"] == "master"] 
                chosen_branch_id = chosen_branch[0]["_id"]
                url = self.base_url + "/branch/" + chosen_branch_id + "/token"
                resp = requests.get(url, headers={'Authorization': 'bearer ' + self.token})
                branch_token = resp.json()['token']
                with open(config.BRANCH_TOKEN_FILE, 'w') as f:
                    f.write(branch_token)
                print("You chose the project {} on the branch master".format(all_projects[int(chosen_project)]["name"]))
                print(branch_token)
            else:    
                print("Wrong input, please type a number between 0 and {}".format(len(all_projects)))


            

    def logout(self):
        if os.path.isfile(config.CREDS_GENERATED_FILE):
            os.remove(config.CREDS_GENERATED_FILE)
        print("Logout successful")    

    def pull_data(self):
        url = self.base_url + '/branch/funcunittests'
        resp = requests.get(url, headers={'Authorization': 'bearer ' + self.token_branch})
        if resp.status_code != 200:
            print("No branch selected, please choose a project: pycodya projects")
            return
        data = resp.json()['unittestdata']
        print("Pulling from Codya")
        tools.create_unitestfiles(data)

    def _send_data(self, data):
        url = self.base_url + '/branch/funcunittest'
        print(data)
        requests.post(url, json=data, headers={'Authorization': 'bearer ' + self.token_branch})





import os
import sys
import json
import time

from pycodya import config

# Add the current path
sys.path.append(os.getcwd())

def import_name(modulename, name):
    """
    Import a named object from a module in the context of this function
   
    Inputs
    :modulename: file you need to import
    :name: name of the function or class that needs to be imported

    Return
    :module: class, function or variables of the file that was imported
    """
    try:
        module = __import__(modulename, globals(), locals(), [name])
        return vars(module)[name]
    except ImportError as error:
        print(error)
        return None


def get_list_files():
    """
    Read list of test files generated from the test

    Return
    :list_files: paths of the list of files 
    """
    if not os.path.isdir(config.DIR_GENERATED_TESTS):
        return []

    list_files = os.listdir(config.DIR_GENERATED_TESTS) 
    return [l for l in list_files if l.endswith('.json')]

def read_file(file_name):
    """
    Extract the data out of the one file

    Input
    :file_name: name of the json file

    Return
    :json_data: json data from the file
    """
    json_data = json.load(open('{}/{}'.format(
        config.DIR_GENERATED_TESTS, 
        file_name
     )))
    return json_data                      



def test_file_functions(file_name, json_data):
    """
    Run all the tests of all the functions for one file

    Input
    :file_name: name of the stored json file 
    :json_data: data for the json file 
    """
    print('\nRunning tests in this file {}'.format(file_name.replace('.json', '')))

    for data in json_data:
        # Extract the name of the module
        module_name = data['filePath'].split('.')[0].replace('/', '.')
        
        # Get the function to be tested
        test_function = import_name(module_name, data['funcName'])
        
        # Make sure that the class CodyaTest is set to True
        file_module = import_name(module_name, config.PYCODYA_VAR_NAME)
        file_module._set_testing(True)

        # Run the test for that function
        input_function = data['dataInput']
        results = test_function(*input_function)

        if len(data['dataOutput']) == 1: data_output = data['dataOutput'][0]
        else: data_output = tuple(data['dataOutput'])

        # Analyse the results
        if results == data_output:
            print('\033[1m file: \033[0m {} - \033[1m function: \033[0m {} - \033[92m test OK \033[0m'
                .format(module_name, data['funcName']))
        else:
            raise AssertionError('\033[91m test FAIL \033[0m {} = {}'.format(results, data))
            

def run_tests():
    """
    Run all the tests from the data stored in the json files
    """
    # Start time to measure how long it takes to run the tests
    start_time = time.time()

    # Get the list of stored files
    list_files = get_list_files()

    if not len(list_files):
        print("No tests stored yet, please see the documentation to run tests")
    else:
        # Run the test for all the files
        for f in list_files:
            json_data = read_file(f)                      
            test_file_functions(f, json_data)

        print("\n\nThe tests took\033[1m {} \033[0m seconds to run".format(time.time() - start_time))






import os
import json
from pycodya import config

def get_relative_path(root_path, file_path):
        root_array = os.path.normpath(root_path).split(os.sep)
        file_array = os.path.normpath(file_path).split(os.sep)
        relative_array = [p for p in file_array if p not in root_array]
        relative_path = os.path.join(*relative_array)
        return relative_path

def create_unitestfiles(data):
    # create folder if not exists
    if not os.path.isdir(config.DIR_GENERATED_TESTS):
        os.makedirs(config.DIR_GENERATED_TESTS)

    # create files

    list_files = []
    for d in data:
        if d['fileName'] not in list_files:
            list_files.append(d['fileName'])


    for file in list_files:
        data_file = []
        for d in data:
            if d['fileName'] == file:
                
                data_file.append({
                    'fileName': d['fileName'],
                    'filePath': d['filePath'],
                    'funcName': d['funcName'],
                    'dataInput': d['dataInput'],
                    'dataOutput': d['dataOutput']
                })
        path_file = os.path.join(
            config.DIR_GENERATED_TESTS,
            file + '.json'
        )
        with open(path_file, 'w') as f:
            json.dump(data_file, f)
        print("Created file: {}".format(path_file))    

        



"""pycodya.

Usage:
  pycodya <actions>  

Options:
  -h --help     Show this screen.
  --version     Show version.
  actions       alltests: test everything


Actions:
 login         Login to the codya
 logout        Remove the session
 projects      Connect to the right project and branch
 alltests      Run all alltests for that project

"""

from docopt import docopt
from pycodya.run_tests import run_tests
from pycodya.codya_api import CodyaApi


def main():
    arguments = docopt(__doc__, version="0.2.1a")

    if arguments['<actions>'] == 'alltests':
        run_tests()
    elif arguments['<actions>'] == 'login':
        CodyaApi().login()
    elif arguments['<actions>'] == 'logout':
        CodyaApi().logout()
    elif arguments['<actions>'] == 'projects':
        CodyaApi().projects()
    elif arguments['<actions>'] == 'pull':
        CodyaApi().pull_data() 
    


if __name__ == '__main__':
    main()
    




import os

config = None
if os.path.exists('pycodya_config'):
    import pycodya_config as config


# Provide the autotest paths
DIR_GENERATED_ROOT = '.pycodya'
DIR_GENERATED_TESTS = os.path.join(DIR_GENERATED_ROOT, 'tests')
CREDS_GENERATED_FILE = os.path.join(DIR_GENERATED_ROOT, 'creds.txt')
BRANCH_TOKEN_FILE = os.path.join(DIR_GENERATED_ROOT, 'branch.txt')

# Provide the variable test
PYCODYA_VAR_NAME = 'CodyaTest' if not config else config.PYCODYA_VAR_NAME



import os
import json
import requests

from pycodya.helpers import tools
from pycodya import config
from pycodya.codya_api import CodyaApi

class Codya(object):
    """
    Class that creates some testing environment for decoration on the local environment
    Only needs to be initated once for the entire project 
    """

    def __init__(self, token=None):
        """
        Initialise the class

        Input
        :token: (uuid) unique to the customer and project
        """
        self.token = token
        self.api = CodyaApi(self.token)
        self.is_testing = False
        self.ROOT_DIR = os.path.abspath(os.curdir)

        if not os.path.isdir(config.DIR_GENERATED_TESTS):
            os.makedirs(config.DIR_GENERATED_TESTS)


    def _set_testing(self, is_testing):
        """
        Setter for is_testing

        Input
        :is_testing: true if we are running the tests
        """
        self.is_testing = is_testing

    def _send_data(self, new_data):
        """
        Send the input and output of the function while the function is runnning
        to codya api server
        
        Inputs
        :new_data: (Obj) new data object with the following keys 
            :file_path: (string) absolute file path of the file which contains the function to be tested
            :function_name: name of the function to be tested
            :data_input: args given to the function to be tested
            :data_output: output of the data to be tested stored in the test files
        """
        self.api._send_data(new_data) 

        
        



    def _get_stored_output(self, file_path, function_name, data_input):
        """
        Get all the stored output from the test files stored while the functions were running

        Inputs
        :file_path: (string) absolute file path of the file which contains the function to be tested
        :function_name: name of the function to be tested
        :data_input: args given to the function to be tested

        Output
        :data_output: output of the data to be tested stored in the test files
        """
        file_name = file_path.split('/')[-1]
        output_file = '{}/{}.json'.format(
            config.DIR_GENERATED_TESTS,
            file_name
        )
        
        if not os.path.isfile(output_file):
            return

        with open(output_file) as fr:
            stored_data = json.load(fr)
        
        relative_path = tools.get_relative_path(self.ROOT_DIR, file_path)

        for d in stored_data:
            if d['filePath'] == relative_path \
               and d['funcName'] == function_name \
               and d['dataInput'] == list(data_input):
                try:
                    return list(d['dataOutput'])
                except:
                    return list([d['dataOnput']])
        return 

    def _store_test(self, file_path, function_name, data_input, data_output, is_mocked=False):
        """
        Store the input and output of the function while the function is runnning
        
        Inputs
        :file_path: (string) absolute file path of the file which contains the function to be tested
        :function_name: name of the function to be tested
        :data_input: args given to the function to be tested
        :data_output: output of the data to be tested stored in the test files
        :is_mocked: check if the function is mocked
        """
        file_name = file_path.split('/')[-1]
        output_file = '{}/{}.json'.format(
            config.DIR_GENERATED_TESTS,
            file_name
        )
        
        if os.path.isfile(output_file):
            with open(output_file) as fr:
                stored_data = json.load(fr)
        else:
            stored_data = []
        
        relative_path = tools.get_relative_path(self.ROOT_DIR, file_path)
        
        try:
            list_data_output = list(data_output)
        except:
            list_data_output = list([data_output])
        
        new_data = {
            'fileName': file_name,    
            'filePath': relative_path, 
            'funcName': function_name,
            'dataInput': list(data_input),
            'dataOutput': list_data_output,
        }
                
        flag = False
        for d in stored_data:
            if not is_mocked and d == new_data:
                flag = True
                break
            if is_mocked \
                and d['filePath'] == new_data['filePath'] \
                and d['funcName'] == new_data['funcName'] \
                and d['dataInput'] == new_data['dataInput']:
                flag = True
                break

        # if the data has not been found, store it
        if not flag:
            stored_data.append(new_data)
            # if there is a token send the data
            if self.token:
                self._send_data(new_data)


        fw = open(output_file, 'w')
        json.dump(stored_data, fw)
        fw.close()

    def function_tester(self, func):
        """
        Function decorator to test some predictable function

        Input
        :func: function to be tested

        Return
        :output: output of the function
        """
        def inner(*args, **kwargs):
            # get the result 
            result = func(*args, **kwargs)

            # store the function attributes for later
            if not self.is_testing:
                self._store_test(
                    func.__globals__['__file__'],
                    func.__name__,
                    args,
                    result,
                    is_mocked=False,
                )
            return result
        return inner
    
    def function_mock(self, func):
        """
        Function decorator to test some unpredictable function.
        During the test time, the function will be mocked

        Input
        :func: function to be tested

        Return
        :output: output of the function
        """
        def inner(*args, **kwargs):
            # 
            if self.is_testing:
                result = self._get_stored_output(
                    func.__globals__['__file__'],
                    func.__name__,
                    args,
                    )
                if len(result) == 1: return result[0]
                else: return tupple(result)

            # get the result 
            result = func(*args, **kwargs)
            
            # store the function attributes for later
            self._store_test(
                func.__globals__['__file__'],
                func.__name__,
                args,
                result,
                is_mocked=True,
            )
            return result 
        return inner


    def class_function_test(self, func):
        def inner(*args, **kwargs):
            result = func(*args, **kwargs)
            print("I can decorate any class", self.token, func.__name__, 'arguments', args[0].__dict__, 'result', result)
            return result
        return inner


